# Package metadata
__author__ = "Sky Christensen"
__author_email__ = "sky@skychristensen.com"
__description__ = "File immutability manager"
__licence__ = "MIT"
__title__ = "Entomb"
__url__ = "https://github.com/countermeasure/entomb"
__version__ = "0.3.1"
