# Command line arguments.
DRY_RUN_ARG = "--dry-run"
DRY_RUN_SHORT_ARG = "-d"
INCLUDE_GIT_ARG = "--include-git"
INCLUDE_GIT_SHORT_ARG = "-g"
LIST_IMMUTABLE_ARG = "--list-immutable"
LIST_MUTABLE_ARG = "--list-mutable"
REPORT_ARG = "--report"
REPORT_SHORT_ARG = "-r"
UNSET_ARG = "--unset"
UNSET_SHORT_ARG = "-u"
VERSION_ARG = "--version"
VERSION_SHORT_ARG = "-v"


# Formatting
TABLE_WIDTH = 40


# Path types.
DIRECTORY = "directory"
FILE = "file"
LINK = "link"
